
"""

Graphical operations on video frames.

We encode and decode videos by forking an mencoder process and
streaming raw video frames in RGB to its standard input and
output. This may not be very fast, but it is simpler than linking
with video libraries.


"""


import os,sys,select,time,random,traceback,popen2

# clear to end of line, nice in combination with \r
clr_eol="\033[K"


##########################################################################################
# I/O on video files
##########################################################################################


class VidPlayer:
  def __init__(self):
    self.t=-1
       
  def get_frame(self):
    """ returns a frame as raw RGB data """
    self.t+=1
    return None

use_popen=True

class FileVidPlayer(VidPlayer):
  
  def get_frame(self):
    """ read the required bytes from file descriptor f """
    self.t+=1

    data=self.f.read(self.w*self.h*3)    
    if not data: return None
    assert len(data)==self.w*self.h*3
    
    return data

  def __del__(self):
    """ need special care because the mencoder process does not
    necessarily exit gracefully on sigpipe"""
    # print "del self=",self
    if use_popen:
      os.kill(self.po.pid,9)
    else:
      pass
      #self.f.close()
      #os.close(self.f.fileno())
    # print "del done"

class VOBPlayer(FileVidPlayer):
  """ Plays a DVD .VOB file and outputs frames in a given output size """

  def __init__(self,fname,skip_secs,w,h):
    FileVidPlayer.__init__(self)
    self.w,self.h=w,h    

    # force aspect ratio
    sw=512*h/288

    cmd="mencoder -quiet -msglevel all=0 \"%s\" -of rawvideo -nosound -ovc raw -vf scale=%d:%d,crop=%d:%d,format=rgb24 -o /dev/stdout -ss %g"%(fname,sw,h,w,h,skip_secs)

    print cmd

    if use_popen:
      self.po=popen2.Popen3(cmd)
      self.f=self.po.fromchild
    else:
      self.f=os.popen(cmd,"r")
    
    for i in range(4):
      # skip 4 uncompressable verbose messages from mencoder
      print self.f.readline(),

    # assert self.get_frame()
    # print fname,"frame 1 ok"


class TrecvidPlayer(FileVidPlayer):

  def __init__(self,fname,skip_secs):
    VidPlayer.__init__(self)

    # first find the input video size

    self.w=self.h=None

    cmd="mplayer -identify -vo null -ao null -frames 0 \"%s\""%fname

    for line in os.popen(cmd,"r"):
      fields=line.split("=")
      if fields[0]=="ID_VIDEO_WIDTH": self.w=int(fields[1])
      if fields[0]=="ID_VIDEO_HEIGHT": self.h=int(fields[1])
      if fields[0]=="ID_VIDEO_FPS": assert float(fields[1])==25.000
    assert self.w and self.h

    cmd="mencoder -quiet -msglevel all=0 \"%s\" -of rawvideo -nosound -ovc raw -vf format=rgb24 -o /dev/stdout -ss %g"%(fname,skip_secs)

    # print cmd

    if use_popen:
      self.po=popen2.Popen3(cmd)
      self.f=self.po.fromchild
    else:
      self.f=os.popen(cmd,"r")
    
    for i in range(4):
      # skip 4 uncompressable verbose messages from mencoder
      print self.f.readline(),

    #assert self.get_frame()
    #print fname,"frame 1 ok",self


class VidSink:
  """ write output video in mpeg1 without sound
  typical bitrates:
  1150 = perfect
  115 = still lookable (youtube)
  64 = getting worse 
  32 = old Realplayer
  4 = Inferno
  """
  
  def __init__(self,fname,w,h,bitrate=1150000):
    self.w=w
    self.h=h
    self.fps=25

    cmd="mencoder -msglevel all=0 -demuxer rawvideo -rawvideo w=%d:h=%d:format=rgb24:fps=%d - -o \"%s\" -of mpeg -ovc lavc -lavcopts vcodec=mpeg1video:vbitrate=%d:keyint=12 -nosound -"%(w,h,self.fps,fname,bitrate)

    # aspect=1.333
    print cmd
    self.f=os.popen(cmd,"w")

  def add_frame(self,data):
    assert len(data)==self.w*self.h*3
    self.f.write(data)

##  def __del__(self):    
##    print "VidSink del"
##    self.f.close()
    


##########################################################################################
# Combinations & video filters
##########################################################################################


class Stage1Video(VidPlayer):

  def __init__(self,trecvid_file,t1,t2,t3,trecvid_secs,other_file=None,other_secs=None):
    """
    Combines 1 or 2 videos into one. 
    0-t1: other, t1-t2: trecvid, t2-t3: other
    other_secs: offset of 0 in other video
    trecvid_secs: offset of t1 in trecvid video    
    """
    VidPlayer.__init__(self)
    self.t1,self.t2,self.t3=t1,t2,t3
    
    self.vid=TrecvidPlayer(trecvid_file,trecvid_secs)
    self.w,self.h=self.vid.w,self.vid.h
    
    if t1!=0 or t2!=t3:
      self.vid2=VOBPlayer(other_file,other_secs,self.vid.w,self.vid.h)
    else:
      self.vid2=None
  
  def get_frame(self):
    self.t+=1
    if self.t>=self.t3:
      return None

    if self.vid2:
      data2=self.vid2.get_frame()

    if self.t1<=self.t<self.t2:
      return self.vid.get_frame()
    else:
      return data2

class FilterVidPlayer(VidPlayer):
  """ video player that filters raw frames """
  
  def __init__(self):
    VidPlayer.__init__(self)
    
  def set_vid(self,vid):    
    self.vid=vid
    self.w,self.h=vid.w,vid.h
  

class TScaler(FilterVidPlayer):
  """ Time scaler
  factor>1.0: faster
  factor<1.0: slower """
  
  def __init__(self,factor):
    FilterVidPlayer.__init__(self)
    self.factor=factor
    
  def get_frame(self):
    # we want vid.t=self.t*factor
    self.t+=1
    while self.vid.t<=self.t*self.factor:
      self.data=self.vid.get_frame()
    return self.data

class FrameDropper(FilterVidPlayer):
  """ replaces somes frames with black ones """

  def __init__(self,proba=0.3):
    FilterVidPlayer.__init__(self)    
    self.proba=proba
    self.tdrop=0
    
  def get_frame(self):
    self.t+=1
    data=self.vid.get_frame()
    if not data: return
    if self.tdrop==0 and random.random()<self.proba:
      self.tdrop=random.randrange(1,6)
    if self.tdrop>0:
      data=len(data)*'\0'
      self.tdrop-=1
    return data



##########################################################################################
# Simple Gimp-based image filters
##########################################################################################

    

from gimpfu import *


class SimpleFilter:
  " filter depending on a single parameter "
  
  def __init__(self,param=0):
    self.param=param

  def filter(self,im,lay):
    " transforms frame stored as a GIMP layer "
    pass


class NoiseFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.plug_in_rgb_noise(im,lay,True,False,self.param,self.param,self.param,0)

class GammaFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.gimp_curves_spline(lay, 0, 6, [0,0,128,255*0.5**self.param,255,255])
  
class ContrastFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.gimp_brightness_contrast(lay,0,self.param)
  
class BlurFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.plug_in_gauss(im,lay, self.param, self.param, 1)
    
class PixelizeFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.plug_in_pixelize(im, lay, self.param)

class ScaleUnscaleFilter(SimpleFilter):
  " blurs frames by scaling them down and scaling up again "
  
  def filter(self,im,lay):
    lay2=pdb.gimp_layer_copy(lay, True)
    pdb.gimp_image_add_layer(im,lay2, -1)
    lay2 = pdb.gimp_drawable_transform_scale_default(
      lay2, 0, 0, im.width/self.param, im.height/self.param, True, 0)
    # nice crappy scale method :-)
    lay2=pdb.gimp_drawable_transform_scale(
      lay2, 0, 0, im.width, im.height, 0, 1, True, 1, 0)

    pdb.gimp_edit_copy(lay2)
    pdb.gimp_image_remove_layer(im,lay2)

    floating_sel=pdb.gimp_edit_paste(lay, False)
    pdb.gimp_floating_sel_anchor(floating_sel)    

    

class MosaicFilter(SimpleFilter):
  " toy " 
  def filter(self,im,lay):
    pdb.plug_in_mosaic(im,lay,self.param,4,1,0.65,True,135,0.2,True,True,1,0,0)

class AspectRatioFilter(SimpleFilter):
  " scales the frames horizontally or vertically, adding black margins "
  
  def filter(self,im,lay):
    lay2=pdb.gimp_layer_copy(lay, True)

    pdb.gimp_image_add_layer(im,lay2, -1)

    if self.param<1:
      pdb.gimp_layer_scale(lay2,im.width,int(im.height*self.param),False)
    else:
      pdb.gimp_layer_scale(lay2,int(im.width/self.param),im.height,False)
      
    pdb.gimp_edit_copy(lay2)

    pdb.gimp_image_remove_layer(im,lay2)
    
    pdb.gimp_context_set_background([0,0,0])
    pdb.gimp_drawable_fill(lay, BACKGROUND_FILL)
    
    floating_sel=pdb.gimp_edit_paste(lay, False)
    pdb.gimp_floating_sel_anchor(floating_sel)    



class BlackBorderFilter(SimpleFilter):
  " shifts the layer " 
  def filter(self,im,lay):
    (dx,dy)=self.param
    pdb.gimp_layer_translate(lay, dx, dy)
    pdb.gimp_context_set_background([0,0,0])
    pdb.gimp_layer_resize_to_image_size(lay)
    # weirdly (fortunately), this does not seem to change the layer

class BlackMargins(SimpleFilter):
  
  def filter(self,im,lay):
    pdb.gimp_rect_select(im,
                         self.param[0],self.param[1],
                         im.width-self.param[2],im.height-self.param[3], 2, False, 0)
    pdb.gimp_selection_invert(im)
    pdb.gimp_context_set_background([0,0,0])
    pdb.gimp_edit_fill(lay, 1)
    
class FlipFilter(SimpleFilter):
  def filter(self,im,lay):
    pdb.gimp_image_flip(im,ORIENTATION_HORIZONTAL)




##########################################################################################
# Complex Gimp-based image filters
##########################################################################################



class OverlayImageFilter:
  " draws a still image over the frames "
  
  def __init__(self,fname):
    self.overlay=pdb.gimp_file_load(fname,fname)

  def __del__(self):
    pdb.gimp_image_delete(self.overlay)

  def filter(self,im,lay):
    pdb.gimp_edit_copy(self.overlay.layers[0])
    floating_sel=pdb.gimp_edit_paste(lay, True)
    pdb.gimp_floating_sel_anchor(floating_sel)

def rescale_and_set_background(im,lay,bglay,x0,y0,x1,y1):
  """ rescale lay to (x0,y0,x1,y1) and add a background (from bglay) to
  it.
  A bit tricky because we don't want to change the drawing layer """
  lay2=pdb.gimp_layer_copy(lay, True)

  pdb.gimp_image_add_layer(im,lay2, -1)

  lay2=pdb.gimp_drawable_transform_scale_default(lay2,x0,y0,x1,y1,True,1)

  pdb.gimp_edit_copy(bglay)

  floating_sel=pdb.gimp_edit_paste(lay, False)
  pdb.gimp_floating_sel_anchor(floating_sel)

  pdb.gimp_edit_copy(lay2)
  pdb.gimp_image_remove_layer(im,lay2)

  floating_sel=pdb.gimp_edit_paste(lay, False)
  pdb.gimp_floating_sel_anchor(floating_sel)    

def rude_color_norm(lay):
  " mimic violent automatic gain ctrl on video camera "
  mean, std_dev, median, pixels, count, percentile = pdb.gimp_histogram(
    lay, HISTOGRAM_VALUE, 0, 256)
  # fact=0.5
  fact=0.7
  curve=[0,0,                                     
         max(mean-std_dev*fact,15),10,
         min(mean+std_dev*fact,240),255,
         255,255]
  #print curv
  pdb.gimp_curves_spline(lay, 0, 8, curve)
  pdb.gimp_curves_spline(lay, HISTOGRAM_BLUE, 6, [0,0,128,140,255,255])
  
  
class CamcordingFilter:
  " emulates camcording the video. Maybe add a BlurFilter(3.0) to improve the effect "
  def __init__(self,coords):
    self.coords=coords
  
  def filter(self,im,lay):
    lay2=pdb.gimp_layer_copy(lay, True)

    pdb.gimp_image_add_layer(im,lay2, -1)

    rude_color_norm(lay2)

    coords=self.coords

    lay2=pdb.gimp_drawable_transform_perspective_default(
      lay2,
      coords[0],coords[1],coords[2],coords[3],
      coords[4],coords[5],coords[6],coords[7],
      True, 1)

    pdb.gimp_edit_copy(lay2)

    pdb.gimp_image_remove_layer(im,lay2)
    
    pdb.gimp_context_set_background([0,0,0])
    pdb.gimp_drawable_fill(lay, BACKGROUND_FILL)
    
    floating_sel=pdb.gimp_edit_paste(lay, False)
    pdb.gimp_floating_sel_anchor(floating_sel)    
    
    

class PictureInVid:
  """ reduce a video to a small window in a background video """
  
  def __init__(self,vid,scale,x0,y0):
    self.scale,self.x0,self.y0=scale,x0,y0
    self.vid=vid

    im=pdb.gimp_image_new(vid.w,vid.h,0)
    self.bglay=pdb.gimp_layer_new(im,vid.w,vid.h,0,"my layer",100, 0)
    
    pdb.gimp_image_add_layer(im,self.bglay,0)
    
    self.pr=self.bglay.get_pixel_rgn(0,0,im.width,im.height,True,False)


  def filter(self,im,lay):

    w=int(im.width*self.scale)
    h=int(im.height*self.scale)
    
    self.pr[0:self.bglay.width,0:self.bglay.height]=self.vid.get_frame()

    rescale_and_set_background(im,lay,self.bglay,self.x0, self.y0,self.x0+w, self.y0+h)

  


##########################################################################################
# Run a filter chain
##########################################################################################



def run_filter(stage1src,filters,sink):
  """
  stage1src should be a VidPlayer
  filters is a list of objects with a filter(im,layer) method
  sink is a VidSink
  """ 
  
  w,h=stage1src.w,stage1src.h
  im=pdb.gimp_image_new(w,h,0)
  lay=pdb.gimp_layer_new(im,w,h,0,"i/o layer",100, 0)

  pdb.gimp_image_add_layer(im,lay,0)

  pr=lay.get_pixel_rgn(0,0,im.width,im.height,True,False)

  print "chain ready"

  t=0
  
  while True:
    data=stage1src.get_frame()

    if not data: break

    print clr_eol,"frame",t,"\r",
    sys.stdout.flush()
    
    pr[0:w,0:h]=data

    for filt in filters:
      filt.filter(im,lay)

    sink.add_frame(pr[0:w,0:h])

    t+=1

  pdb.gimp_image_delete(im)
  print "chain done"

